@@ -231,6 +231,10 @@ class ShellTest(utils.TestCase):
         # NOTE(jdg): we default to detail currently
         self.assert_called('GET', '/volumes/detail')
 
+    def test_list_with_with_count(self):
+        self.run_command('--os-volume-api-version 3.45 list --with-count')
+        self.assert_called('GET', '/volumes/detail?with_count=True')
+
     def test_summary(self):
         self.run_command('--os-volume-api-version 3.12 summary')
         self.assert_called('GET', '/volumes/summary')
@@ -240,6 +244,20 @@ class ShellTest(utils.TestCase):
                           self.run_command,
                           'list --group_id fake_id')
 
+    def test_type_list_with_filters_invalid(self):
+        self.assertRaises(exceptions.UnsupportedAttribute,
+                          self.run_command,
+                          '--os-volume-api-version 3.51 type-list '
+                          '--filters key=value')
+
+    def test_type_list_with_filters(self):
+        self.run_command('--os-volume-api-version 3.52 type-list '
+                         '--filters extra_specs={key:value}')
+        self.assert_called(
+            'GET', '/types?%s' % parse.urlencode(
+                {'extra_specs':
+                    {six.text_type('key'): six.text_type('value')}}))
+
     @ddt.data("3.10", "3.11")
     def test_list_with_group_id_after_3_10(self, version):
         command = ('--os-volume-api-version %s list --group_id fake_id' %
@@ -376,6 +394,12 @@ class ShellTest(utils.TestCase):
         self.assert_called('PUT', '/attachments/1234', body={'attachment':
                                                              body})
 
+    @ddt.unpack
+    def test_attachment_complete(self):
+        command = '--os-volume-api-version 3.44 attachment-complete 1234'
+        self.run_command(command)
+        self.assert_called('POST', '/attachments/1234/action', body=None)
+
     def test_attachment_delete(self):
         self.run_command('--os-volume-api-version 3.27 '
                          'attachment-delete 1234')
@@ -424,20 +448,36 @@ class ShellTest(utils.TestCase):
         expected = {'backup': {'name': 'new_name'}}
         self.assert_called('PUT', '/backups/1234', body=expected)
 
+    def test_backup_list_with_with_count(self):
+        self.run_command(
+            '--os-volume-api-version 3.45 backup-list --with-count')
+        self.assert_called('GET', '/backups/detail?with_count=True')
+
     def test_backup_update_with_description(self):
         self.run_command('--os-volume-api-version 3.9 '
                          'backup-update 1234 --description=new-description')
         expected = {'backup': {'description': 'new-description'}}
         self.assert_called('PUT', '/backups/1234', body=expected)
 
+    def test_backup_update_with_metadata(self):
+        cmd = '--os-volume-api-version 3.43 '
+        cmd += 'backup-update '
+        cmd += '--metadata foo=bar '
+        cmd += '1234'
+        self.run_command(cmd)
+        expected = {'backup': {'metadata': {'foo': 'bar'}}}
+        self.assert_called('PUT', '/backups/1234', body=expected)
+
     def test_backup_update_all(self):
         # rename and change description
-        self.run_command('--os-volume-api-version 3.9 '
+        self.run_command('--os-volume-api-version 3.43 '
                          'backup-update --name new-name '
-                         '--description=new-description 1234')
+                         '--description=new-description '
+                         '--metadata foo=bar 1234')
         expected = {'backup': {
             'name': 'new-name',
             'description': 'new-description',
+            'metadata': {'foo': 'bar'}
         }}
         self.assert_called('PUT', '/backups/1234', body=expected)
 
@@ -515,9 +555,39 @@ class ShellTest(utils.TestCase):
                                'metadata': {},
                                'volume_type': '4321',
                                'description': None,
-                               'multiattach': False}}
+                               'multiattach': False,
+                               'backup_id': None}}
         self.assert_called_anytime('POST', '/volumes', expected)
 
+    @ddt.data({'cmd': '--os-volume-api-version 3.47 create --backup-id 1234',
+               'update': {'backup_id': '1234'}},
+              {'cmd': '--os-volume-api-version 3.47 create 2',
+               'update': {'size': 2}}
+              )
+    @ddt.unpack
+    def test_create_volume_with_backup(self, cmd, update):
+        self.run_command(cmd)
+        self.assert_called('GET', '/volumes/1234')
+        expected = {'volume': {'imageRef': None,
+                               'project_id': None,
+                               'status': 'creating',
+                               'user_id': None,
+                               'size': None,
+                               'availability_zone': None,
+                               'source_replica': None,
+                               'attach_status': 'detached',
+                               'source_volid': None,
+                               'consistencygroup_id': None,
+                               'name': None,
+                               'snapshot_id': None,
+                               'metadata': {},
+                               'volume_type': None,
+                               'description': None,
+                               'multiattach': False,
+                               'backup_id': None}}
+        expected['volume'].update(update)
+        self.assert_called_anytime('POST', '/volumes', body=expected)
+
     def test_group_list(self):
         self.run_command('--os-volume-api-version 3.13 group-list')
         self.assert_called_anytime('GET', '/groups/detail')
@@ -549,9 +619,6 @@ class ShellTest(utils.TestCase):
     def test_group_create(self):
         expected = {'group': {'name': 'test-1',
                               'description': 'test-1-desc',
-                              'user_id': None,
-                              'project_id': None,
-                              'status': 'creating',
                               'group_type': 'my_group_type',
                               'volume_types': ['type1', 'type2'],
                               'availability_zone': 'zone1'}}
@@ -597,10 +664,7 @@ class ShellTest(utils.TestCase):
     def test_group_snapshot_create(self):
         expected = {'group_snapshot': {'name': 'test-1',
                                        'description': 'test-1-desc',
-                                       'user_id': None,
-                                       'project_id': None,
-                                       'group_id': '1234',
-                                       'status': 'creating'}}
+                                       'group_id': '1234'}}
         self.run_command('--os-volume-api-version 3.14 '
                          'group-snapshot-create --name test-1 '
                          '--description test-1-desc 1234')
@@ -616,9 +680,6 @@ class ShellTest(utils.TestCase):
     def test_group_create_from_src(self, grp_snap_id, src_grp_id, src):
         expected = {'create-from-src': {'name': 'test-1',
                                         'description': 'test-1-desc',
-                                        'user_id': None,
-                                        'project_id': None,
-                                        'status': 'creating',
                                         'group_snapshot_id': grp_snap_id,
                                         'source_group_id': src_grp_id}}
         cmd = ('--os-volume-api-version 3.14 '
@@ -643,6 +704,11 @@ class ShellTest(utils.TestCase):
                          'manageable-list fakehost --detailed False')
         self.assert_called('GET', '/manageable_volumes?host=fakehost')
 
+    def test_volume_manageable_list_cluster(self):
+        self.run_command('--os-volume-api-version 3.17 '
+                         'manageable-list --cluster dest')
+        self.assert_called('GET', '/manageable_volumes/detail?cluster=dest')
+
     def test_snapshot_manageable_list(self):
         self.run_command('--os-volume-api-version 3.8 '
                          'snapshot-manageable-list fakehost')
@@ -658,6 +724,36 @@ class ShellTest(utils.TestCase):
                          'snapshot-manageable-list fakehost --detailed False')
         self.assert_called('GET', '/manageable_snapshots?host=fakehost')
 
+    def test_snapshot_manageable_list_cluster(self):
+        self.run_command('--os-volume-api-version 3.17 '
+                         'snapshot-manageable-list --cluster dest')
+        self.assert_called('GET', '/manageable_snapshots/detail?cluster=dest')
+
+    @ddt.data('', 'snapshot-')
+    def test_manageable_list_cluster_before_3_17(self, prefix):
+        self.assertRaises(exceptions.UnsupportedAttribute,
+                          self.run_command,
+                          '--os-volume-api-version 3.16 '
+                         '%smanageable-list --cluster dest' % prefix)
+
+    @mock.patch('cinderclient.shell.CinderClientArgumentParser.error')
+    @ddt.data('', 'snapshot-')
+    def test_manageable_list_mutual_exclusion(self, prefix, error_mock):
+        error_mock.side_effect = SystemExit
+        self.assertRaises(SystemExit,
+                          self.run_command,
+                          '--os-volume-api-version 3.17 '
+                         '%smanageable-list fakehost --cluster dest' % prefix)
+
+    @mock.patch('cinderclient.shell.CinderClientArgumentParser.error')
+    @ddt.data('', 'snapshot-')
+    def test_manageable_list_missing_required(self, prefix, error_mock):
+        error_mock.side_effect = SystemExit
+        self.assertRaises(SystemExit,
+                          self.run_command,
+                          '--os-volume-api-version 3.17 '
+                         '%smanageable-list' % prefix)
+
     def test_list_messages(self):
         self.run_command('--os-volume-api-version 3.3 message-list')
         self.assert_called('GET', '/messages')
@@ -736,6 +832,11 @@ class ShellTest(utils.TestCase):
         expected = {'os-reset_status': expected}
         self.assert_called('POST', '/volumes/1234/action', body=expected)
 
+    def test_snapshot_list_with_with_count(self):
+        self.run_command(
+            '--os-volume-api-version 3.45 snapshot-list --with-count')
+        self.assert_called('GET', '/snapshots/detail?with_count=True')
+
     def test_snapshot_list_with_metadata(self):
         self.run_command('--os-volume-api-version 3.22 '
                          'snapshot-list --metadata key1=val1')
@@ -990,3 +1091,199 @@ class ShellTest(utils.TestCase):
         self.assertEqual([mock.call(poll_period)] * 2,
                 mock_time.sleep.call_args_list)
         self.assertEqual([mock.call(some_id)] * 2, poll_fn.call_args_list)
+
+    def test_backup(self):
+        self.run_command('--os-volume-api-version 3.42 backup-create '
+                         '--name 1234 1234')
+        expected = {'backup': {'volume_id': 1234,
+                               'container': None,
+                               'name': '1234',
+                               'description': None,
+                               'incremental': False,
+                               'force': False,
+                               'snapshot_id': None,
+                               }}
+        self.assert_called('POST', '/backups', body=expected)
+
+    def test_backup_with_metadata(self):
+        self.run_command('--os-volume-api-version 3.43 backup-create '
+                         '--metadata foo=bar --name 1234 1234')
+        expected = {'backup': {'volume_id': 1234,
+                               'container': None,
+                               'name': '1234',
+                               'description': None,
+                               'incremental': False,
+                               'force': False,
+                               'snapshot_id': None,
+                               'metadata': {'foo': 'bar'}, }}
+        self.assert_called('POST', '/backups', body=expected)
+
+    def test_backup_with_az(self):
+        self.run_command('--os-volume-api-version 3.51 backup-create '
+                         '--availability-zone AZ2 --name 1234 1234')
+        expected = {'backup': {'volume_id': 1234,
+                               'container': None,
+                               'name': '1234',
+                               'description': None,
+                               'incremental': False,
+                               'force': False,
+                               'snapshot_id': None,
+                               'availability_zone': 'AZ2'}}
+        self.assert_called('POST', '/backups', body=expected)
+
+    @mock.patch("cinderclient.utils.print_list")
+    def test_snapshot_list_with_userid(self, mock_print_list):
+        """Ensure 3.41 provides User ID header."""
+        self.run_command('--os-volume-api-version 3.41 snapshot-list')
+        self.assert_called('GET', '/snapshots/detail')
+        columns = ['ID', 'Volume ID', 'Status', 'Name', 'Size', 'User ID']
+        mock_print_list.assert_called_once_with(mock.ANY, columns,
+                                                sortby_index=0)
+
+    @mock.patch('cinderclient.v3.volumes.Volume.migrate_volume')
+    def test_migrate_volume_before_3_16(self, v3_migrate_mock):
+        self.run_command('--os-volume-api-version 3.15 '
+                         'migrate 1234 fakehost')
+
+        v3_migrate_mock.assert_called_once_with(
+            'fakehost', False, False, None)
+
+    @mock.patch('cinderclient.v3.volumes.Volume.migrate_volume')
+    def test_migrate_volume_3_16(self, v3_migrate_mock):
+        self.run_command('--os-volume-api-version 3.16 '
+                         'migrate 1234 fakehost')
+        self.assertEqual(4, len(v3_migrate_mock.call_args[0]))
+
+    def test_migrate_volume_with_cluster_before_3_16(self):
+        self.assertRaises(exceptions.UnsupportedAttribute,
+                          self.run_command,
+                          '--os-volume-api-version 3.15 '
+                          'migrate 1234 fakehost --cluster fakecluster')
+
+    @mock.patch('cinderclient.shell.CinderClientArgumentParser.error')
+    def test_migrate_volume_mutual_exclusion(self, error_mock):
+        error_mock.side_effect = SystemExit
+        self.assertRaises(SystemExit,
+                          self.run_command,
+                          '--os-volume-api-version 3.16 '
+                          'migrate 1234 fakehost --cluster fakecluster')
+        msg = 'argument --cluster: not allowed with argument <host>'
+        error_mock.assert_called_once_with(msg)
+
+    @mock.patch('cinderclient.shell.CinderClientArgumentParser.error')
+    def test_migrate_volume_missing_required(self, error_mock):
+        error_mock.side_effect = SystemExit
+        self.assertRaises(SystemExit,
+                          self.run_command,
+                          '--os-volume-api-version 3.16 '
+                          'migrate 1234')
+        msg = 'one of the arguments <host> --cluster is required'
+        error_mock.assert_called_once_with(msg)
+
+    def test_migrate_volume_host(self):
+        self.run_command('--os-volume-api-version 3.16 '
+                         'migrate 1234 fakehost')
+        expected = {'os-migrate_volume': {'force_host_copy': False,
+                                          'lock_volume': False,
+                                          'host': 'fakehost'}}
+        self.assert_called('POST', '/volumes/1234/action', body=expected)
+
+    def test_migrate_volume_cluster(self):
+        self.run_command('--os-volume-api-version 3.16 '
+                         'migrate 1234 --cluster mycluster')
+        expected = {'os-migrate_volume': {'force_host_copy': False,
+                                          'lock_volume': False,
+                                          'cluster': 'mycluster'}}
+        self.assert_called('POST', '/volumes/1234/action', body=expected)
+
+    def test_migrate_volume_bool_force(self):
+        self.run_command('--os-volume-api-version 3.16 '
+                         'migrate 1234 fakehost --force-host-copy '
+                         '--lock-volume')
+        expected = {'os-migrate_volume': {'force_host_copy': True,
+                                          'lock_volume': True,
+                                          'host': 'fakehost'}}
+        self.assert_called('POST', '/volumes/1234/action', body=expected)
+
+    def test_migrate_volume_bool_force_false(self):
+        # Set both --force-host-copy and --lock-volume to False.
+        self.run_command('--os-volume-api-version 3.16 '
+                         'migrate 1234 fakehost --force-host-copy=False '
+                         '--lock-volume=False')
+        expected = {'os-migrate_volume': {'force_host_copy': 'False',
+                                          'lock_volume': 'False',
+                                          'host': 'fakehost'}}
+        self.assert_called('POST', '/volumes/1234/action', body=expected)
+
+        # Do not set the values to --force-host-copy and --lock-volume.
+        self.run_command('--os-volume-api-version 3.16 '
+                         'migrate 1234 fakehost')
+        expected = {'os-migrate_volume': {'force_host_copy': False,
+                                          'lock_volume': False,
+                                          'host': 'fakehost'}}
+        self.assert_called('POST', '/volumes/1234/action',
+                           body=expected)
+
+    @ddt.data({'bootable': False, 'by_id': False, 'cluster': None},
+              {'bootable': True, 'by_id': False, 'cluster': None},
+              {'bootable': False, 'by_id': True, 'cluster': None},
+              {'bootable': True, 'by_id': True, 'cluster': None},
+              {'bootable': True, 'by_id': True, 'cluster': 'clustername'})
+    @ddt.unpack
+    def test_volume_manage(self, bootable, by_id, cluster):
+        cmd = ('--os-volume-api-version 3.16 '
+               'manage host1 some_fake_name --name foo --description bar '
+               '--volume-type baz --availability-zone az '
+               '--metadata k1=v1 k2=v2')
+        if by_id:
+            cmd += ' --id-type source-id'
+        if bootable:
+            cmd += ' --bootable'
+        if cluster:
+            cmd += ' --cluster ' + cluster
+
+        self.run_command(cmd)
+        ref = 'source-id' if by_id else 'source-name'
+        expected = {'volume': {'host': 'host1',
+                               'ref': {ref: 'some_fake_name'},
+                               'name': 'foo',
+                               'description': 'bar',
+                               'volume_type': 'baz',
+                               'availability_zone': 'az',
+                               'metadata': {'k1': 'v1', 'k2': 'v2'},
+                               'bootable': bootable}}
+        if cluster:
+            expected['volume']['cluster'] = cluster
+        self.assert_called_anytime('POST', '/os-volume-manage', body=expected)
+
+    def test_volume_manage_before_3_16(self):
+        """Cluster optional argument was not acceptable."""
+        self.assertRaises(exceptions.UnsupportedAttribute,
+                          self.run_command,
+                          'manage host1 some_fake_name '
+                          '--cluster clustername'
+                          '--name foo --description bar --bootable '
+                          '--volume-type baz --availability-zone az '
+                          '--metadata k1=v1 k2=v2')
+
+    def test_worker_cleanup_before_3_24(self):
+        self.assertRaises(SystemExit,
+                          self.run_command,
+                          'work-cleanup fakehost')
+
+    def test_worker_cleanup(self):
+        self.run_command('--os-volume-api-version 3.24 '
+                         'work-cleanup --cluster clustername --host hostname '
+                         '--binary binaryname --is-up false --disabled true '
+                         '--resource-id uuid --resource-type Volume '
+                         '--service-id 1')
+        expected = {'cluster_name': 'clustername',
+                    'host': 'hostname',
+                    'binary': 'binaryname',
+                    'is_up': 'false',
+                    'disabled': 'true',
+                    'resource_id': 'uuid',
+                    'resource_type': 'Volume',
+                    'service_id': 1}
+
+        self.assert_called('POST', '/workers/cleanup', body=expected)
